/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * vim: set ts=8 sts=4 et sw=4 tw=99:
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#include "jit/Bailouts.h"
#include "jit/BaselineFrame.h"
#include "jit/CalleeToken.h"
#include "jit/JitFrames.h"
#include "jit/JitRuntime.h"
#ifdef JS_ION_PERF
#  include "jit/PerfSpewer.h"
#endif
#include "jit/ppc64/SharedICRegisters-ppc64.h"
#include "jit/VMFunctions.h"
#include "vm/JitActivation.h"  // js::jit::JitActivation
#include "vm/JSContext.h"

#include "jit/MacroAssembler-inl.h"

#if DEBUG

/* Useful class to print visual guard blocks. */
class TrampolineAutoDeBlock
{
    private:
        const char *blockname;

    public:
        TrampolineAutoDeBlock(const char *name) {
            blockname = name;
            JitSpew(JitSpew_Codegen, "[[[[[[[[ Trampoline: %s", blockname);
        }

        ~TrampolineAutoDeBlock() {
            JitSpew(JitSpew_Codegen, "         Trampoline: %s ]]]]]]]]", blockname);
        }
};

#undef ADBlock
#define ADBlock(x)  TrampolineAutoDeBlock _adbx(x)
            
#else

/* Useful preprocessor macros to completely elide visual guard blocks. */

#undef ADBlock
#define ADBlock(x)  ;

#endif

using namespace js;
using namespace js::jit;

// All registers to save and restore. This includes the stack pointer, since we
// use the ability to reference register values on the stack by index.
static const LiveRegisterSet AllRegs =
    LiveRegisterSet(GeneralRegisterSet(Registers::AllMask),
                         FloatRegisterSet(FloatRegisters::AllMask));

// This trampoline adheres to the PowerPC ELF64 ABI.
static_assert(sizeof(uintptr_t) == sizeof(uint64_t), "Not 32-bit clean.");

/*

From http://refspecs.linuxfoundation.org/ELF/ppc64/PPC-elf64abi.html#FUNC-CALL :

"Registers r1 [SP], r14 through r31, and f14 through f31 are nonvolatile."
"The condition code register fields CR0, CR1, CR5, CR6, and CR7 are volatile."
"Registers r0, r3 through r12, f0 through f13, and the special purpose registers LR, CTR,
XER, and FPSCR are volatile."
"Furthermore, registers r0, r2, r11, and r12 may be modified by cross-module calls."

Figure 3-17. Stack Frame Organization
(marked up with our values)

High Address

          +-> Back chain (SP + 416)
          |   Floating point register save area (SP + 272) (18dw here for f14-f31)
          |   General register save area (SP + 128) (18dw here for r14-r31)
          |   VRSAVE save word (32 bits) (SP + 124)
          |   Alignment padding (4 or 12 bytes) (SP + 112) (12b here)
          |   Vector register save area (quadword aligned) (SP + 112) (not used)
          |   Local variable space   (SP + 112) (not used)
          |   Parameter save area    (SP + 48) (fixed size minimum 8dw here)
          |   TOC save area          (SP + 40)
          |   link editor doubleword (SP + 32)
          |   compiler doubleword    (SP + 24)
          |   LR save area           (SP + 16)
          |   CR save area           (SP + 8)
SP  --->  +-- Back chain             (SP + 0)

Low Address

"The stack pointer shall maintain quadword alignment."

*/

struct EnterJITRegs
{
// SP + 0
    uint64_t sp;
    uint64_t cr;
    uint64_t lr;

    uint64_t comp;
    uint64_t le;
    uint64_t toc;

    uint64_t parameters[7];

// SP + 112
// We cheat here: this is a safe place to store our old VP.
// In the ABI definition this is just padding.
    uint64_t savedvp;
    uint32_t padding;
    uint32_t vrsave;

// SP + 128
    uint64_t r14;
    uint64_t r15;
    uint64_t r16;
    uint64_t r17;
    uint64_t r18;
    uint64_t r19;
    uint64_t r20;
    uint64_t r21;
    uint64_t r22;
    uint64_t r23;
    uint64_t r24;
    uint64_t r25;
    uint64_t r26;
    uint64_t r27;
    uint64_t r28;
    uint64_t r29;
    uint64_t r30;
    uint64_t r31;

    double f14;
    double f15;
    double f16;
    double f17;
    double f18;
    double f19;
    double f20;
    double f21;
    double f22;
    double f23;
    double f24;
    double f25;
    double f26;
    double f27;
    double f28;
    double f29;
    double f30;
    double f31;

    uint64_t mo_paddin; // make size 16-byte
};

static_assert(sizeof(EnterJITRegs) == 416, "Unexpected size of register save frame");
static_assert(offsetof(EnterJITRegs, sp) == 0, "Register save frame is incorrectly oriented");

// Generates a trampoline for calling JIT code from a C++ function.
// The signature is
// EnterJitCode(void* code, unsigned argc, Value* argv, InterpreterFrame* fp,
//              CalleeToken calleeToken, JSObject* envChain,
//              size_t numStackValues, Value* vp);
// Happily, this all fits into registers.
void
JitRuntime::generateEnterJIT(JSContext* cx, MacroAssembler& masm)
{
    AutoCreatedBy acb(masm, "JitRuntime::generateEnterJIT");
    ADBlock("generateEnterJIT");

    enterJITOffset_ = startTrampolineCode(masm);

    const Register reg_code = IntArgReg0; // r3
    const Register reg_argc = IntArgReg1; // r4
    const Register reg_argv = IntArgReg2; // r5
    const mozilla::DebugOnly<Register> reg_frame = IntArgReg3; // r6
    const Register reg_token = IntArgReg4; // r7
    const Register reg_chain = IntArgReg5; // r8
    const Register reg_values = IntArgReg6; // r9
    const Register reg_vp = IntArgReg7; // r10

    MOZ_ASSERT(OsrFrameReg == reg_frame);

    // Standard Power prologue, more or less.
    // First save LR and CR to the caller linkage area.
    masm.xs_mflr(ScratchRegister);
    masm.as_std(ScratchRegister, StackPointer, offsetof(EnterJITRegs, lr)); // caller
    masm.as_mfcr(ScratchRegister);
    masm.as_std(ScratchRegister, StackPointer, offsetof(EnterJITRegs, cr)); // caller
    // Save SP to the caller linkage area and pull down the new frame.
    masm.as_stdu(StackPointer, StackPointer, -((uint16_t)sizeof(EnterJITRegs)));

    // Save non-volatile registers.
#define SAVE(x) masm.as_std(x, StackPointer, offsetof(EnterJITRegs, x));
    SAVE(r14)
    SAVE(r15)
    SAVE(r16)
    SAVE(r17)
    SAVE(r18)
    SAVE(r19)
    SAVE(r20)
    SAVE(r21)
    SAVE(r22)
    SAVE(r23)
    SAVE(r24)
    SAVE(r25)
    SAVE(r26)
    SAVE(r27)
    SAVE(r28)
    SAVE(r29)
    SAVE(r30)
    SAVE(r31)
#undef SAVE
#define SAVE(x) masm.as_stfd(x, StackPointer, offsetof(EnterJITRegs, x));
    SAVE(f14)
    SAVE(f15)
    SAVE(f16)
    SAVE(f17)
    SAVE(f18)
    SAVE(f19)
    SAVE(f20)
    SAVE(f21)
    SAVE(f22)
    SAVE(f23)
    SAVE(f24)
    SAVE(f25)
    SAVE(f26)
    SAVE(f27)
    SAVE(f28)
    SAVE(f29)
    SAVE(f30)
    SAVE(f31)
#undef SAVE

    // Save VP for the end.
    // We would also save VRSAVE here, if we were likely to use VMX/VSX.
    // We load nargs a little later.
    masm.as_std(reg_vp, StackPointer, offsetof(EnterJITRegs, savedvp)); 

    // Save stack pointer as baseline frame.
    masm.movePtr(StackPointer, FramePointer);

    /***************************************************************
    Loop over argv vector, push arguments onto stack in reverse order
    ***************************************************************/

    // If we are constructing, the count also needs to include newTarget.
    MOZ_ASSERT(CalleeToken_FunctionConstructing == 0x01);
    masm.as_andi_rc(ScratchRegister, reg_token, CalleeToken_FunctionConstructing);
    masm.as_add(reg_argc, reg_argc, ScratchRegister);

    // |Value| is 8-byte aligned, but we want to maintain 16-byte alignment,
    // so tack on an extra Value if the number of arguments is odd.
    // Set the address to copy from to *vp + (argc * 8).
    // WARNING: ABI compliant stack frames are now no longer guaranteed.
    MOZ_ASSERT(sizeof(Value) == 8);
    masm.as_andi_rc(ScratchRegister, reg_argc, 1);
    masm.x_slwi(SecondScratchReg, reg_argc, 3); // times 8
    masm.x_slwi(ScratchRegister, ScratchRegister, 3); // times 8
    masm.addPtr(reg_argv, SecondScratchReg);
    masm.as_add(StackPointer, StackPointer, ScratchRegister);

    // Loop over arguments, copying them from the JS buffer onto the Ion
    // stack so they can be accessed from JIT'ed code.
    Label header, footer;
    // If there aren't any arguments, don't do anything.
    masm.ma_bc(SecondScratchReg, reg_argv, &footer, Assembler::BelowOrEqual, ShortJump);
    {
        masm.bind(&header);

        masm.subPtr(Imm32(sizeof(Value)), SecondScratchReg);
        masm.subPtr(Imm32(sizeof(Value)), StackPointer);

        masm.as_ld(ScratchRegister, SecondScratchReg, 0);
        // XXX: Is this usually on stack? Would inserting nops here help?
        masm.as_std(ScratchRegister, StackPointer, 0);

// XXX: bdnz
        masm.ma_bc(SecondScratchReg, reg_argv, &header, Assembler::Above, ShortJump);
    }
    masm.bind(&footer);

    // Load the number of actual arguments (a 32-bit word), then push the
    // callee token and actual arguments as part of the new frame.
    masm.push(reg_token);
    masm.as_lwz(ScratchRegister, reg_vp, 0);
    masm.pushFrameDescriptorForJitCall(FrameType::CppToJSJit,
      ScratchRegister, ScratchRegister);

    CodeLabel returnLabel;
    Label oomReturnLabel;
    {
        // Handle Interpreter -> Baseline OSR.
        AllocatableGeneralRegisterSet regs(GeneralRegisterSet::All());
        MOZ_ASSERT(!regs.has(FramePointer));
        regs.take(OsrFrameReg);
        regs.take(reg_code);
        MOZ_ASSERT(reg_code == ReturnReg); // regs.take(ReturnReg);
        MOZ_ASSERT(!regs.has(ReturnReg), "ReturnReg matches reg_code");
#if(0)
        // On Power reg_code and the ReturnReg are always aliased because of
        // ABI requirements. The first argument passed, the code pointer,
        // comes in r3, and the ABI requires that r3 be the return register.
        // Therefore, we don't implement the changes in bug 1770922.
        regs.take(JSReturnOperand); // ???
#endif

        Label notOsr;
        masm.ma_bc(OsrFrameReg, OsrFrameReg, &notOsr, Assembler::Zero, ShortJump);

        Register numStackValues = reg_values;
        regs.take(numStackValues);
        Register scratch = regs.takeAny();

        // Frame prologue.
        // Push return address.
        masm.subPtr(Imm32(sizeof(uintptr_t) * 2), StackPointer);
        masm.ma_li(scratch, &returnLabel);
        masm.storePtr(scratch, Address(StackPointer, sizeof(uintptr_t)));
        // Push previous frame pointer. Recovered at frame epilogue.
        masm.storePtr(FramePointer, Address(StackPointer, 0));

        // Reserve frame.
        masm.movePtr(StackPointer, FramePointer);
        masm.subPtr(Imm32(BaselineFrame::Size()), StackPointer);

        Register framePtrScratch = regs.takeAny();
        masm.movePtr(StackPointer, framePtrScratch);

        // Reserve space for locals and stack values.
        masm.x_sldi(scratch, numStackValues, 3);
        masm.subPtr(scratch, StackPointer);

        // Enter exit frame.
        masm.reserveStack(3 * sizeof(uintptr_t));
        masm.storePtr(
            ImmWord(MakeFrameDescriptor(FrameType::BaselineJS)),
            Address(StackPointer, 2 * sizeof(uintptr_t)));  // Frame descriptor
        masm.storePtr(scratch,
            Address(StackPointer, sizeof(uintptr_t))); // fake return address
        masm.storePtr(FramePointer, Address(StackPointer, 0));

        // No GC things to mark, so push a bare token.
        masm.loadJSContext(scratch);
        masm.enterFakeExitFrame(scratch, scratch, ExitFrameType::Bare);

        masm.reserveStack(2 * sizeof(uintptr_t));
        masm.storePtr(FramePointer,
            Address(StackPointer, sizeof(uintptr_t))); // BaselineFrame
        masm.storePtr(reg_code, Address(StackPointer, 0)); // jitcode

        using Fn = bool (*)(BaselineFrame * frame, InterpreterFrame * interpFrame,
                            uint32_t numStackValues);
        masm.setupUnalignedABICall(scratch);
        masm.passABIArg(framePtrScratch); // BaselineFrame
        masm.passABIArg(OsrFrameReg); // InterpreterFrame
        masm.passABIArg(numStackValues);
        masm.callWithABI<Fn, jit::InitBaselineFrameForOsr>(
            ABIType::General, CheckUnsafeCallWithABI::DontCheckHasExitFrame);

        regs.add(OsrFrameReg);
        Register jitcode = regs.takeAny();
        masm.loadPtr(Address(StackPointer, 0), jitcode);
        masm.loadPtr(Address(StackPointer, sizeof(uintptr_t)), FramePointer);
        masm.freeStack(2 * sizeof(uintptr_t));

        Label error;
        masm.freeStack(ExitFrameLayout::SizeWithFooter());
        masm.branchIfFalseBool(ReturnReg, &error);

        // If OSR-ing, then emit instrumentation for setting lastProfilerFrame
        // if profiler instrumentation is enabled.
        {
            Label skipProfilingInstrumentation;
            AbsoluteAddress addressOfEnabled(cx->runtime()->geckoProfiler().addressOfEnabled());
            masm.branch32(Assembler::Equal, addressOfEnabled, Imm32(0),
                          &skipProfilingInstrumentation);
            masm.profilerEnterFrame(FramePointer, scratch);
            masm.bind(&skipProfilingInstrumentation);
        }

//masm.xs_trap_tagged(Assembler::DebugTag0);
        masm.jump(jitcode);

        // OOM: frame epilogue, load error value, discard return address
        // and return.
        masm.bind(&error);
        masm.movePtr(FramePointer, StackPointer); // don't need to reload FP
        masm.addPtr(Imm32(2 * sizeof(uintptr_t)), StackPointer);
        masm.moveValue(MagicValue(JS_ION_ERROR), JSReturnOperand);
        masm.jump(&oomReturnLabel);

        masm.bind(&notOsr);
        // Load the scope chain in R1.
        MOZ_ASSERT(R1.scratchReg() != reg_code);
        masm.ma_move(R1.scratchReg(), reg_chain);
    }

    // The call will push the return address on the stack, thus we check that
    // the stack would be aligned once the call is complete.
// XXX: I don't think this is appropriate for us. We don't claim to be ABI
// compliant at this point, and we pass all the stack invariants.
//    masm.assertStackAlignment(JitStackAlignment, 16);

    // Call the function with pushing return address to stack.
//masm.xs_trap_tagged(Assembler::DebugTag0);
    masm.callJitNoProfiler(reg_code);

    {
        // Interpreter -> Baseline OSR will return here.
        masm.bind(&returnLabel);
        masm.addCodeLabel(returnLabel);
        masm.bind(&oomReturnLabel);
    }

    // Discard arguments and padding. Set sp to the address of the EnterJITRegs
    // on the stack.
    masm.mov(FramePointer, StackPointer);

    // Store the returned value into the vp.
    masm.as_ld(reg_vp, StackPointer, offsetof(EnterJITRegs, savedvp));
    masm.storeValue(JSReturnOperand, Address(reg_vp, 0));

    // Restore non-volatile registers and return.
    // Standard PowerPC epilogue, more or less.
    // Load registers.
#define LOAD(x) masm.as_ld(x, StackPointer, offsetof(EnterJITRegs, x));
    LOAD(r14)
    LOAD(r15)
    LOAD(r16)
    LOAD(r17)
    LOAD(r18)
    LOAD(r19)
    LOAD(r20)
    LOAD(r21)
    LOAD(r22)
    LOAD(r23)
    LOAD(r24)
    LOAD(r25)
    LOAD(r26)
    LOAD(r27)
    LOAD(r28)
    LOAD(r29)
    LOAD(r30)
    LOAD(r31)
#undef LOAD
#define LOAD(x) masm.as_lfd(x, StackPointer, offsetof(EnterJITRegs, x));
    LOAD(f14)
    LOAD(f15)
    LOAD(f16)
    LOAD(f17)
    LOAD(f18)
    LOAD(f19)
    LOAD(f20)
    LOAD(f21)
    LOAD(f22)
    LOAD(f23)
    LOAD(f24)
    LOAD(f25)
    LOAD(f26)
    LOAD(f27)
    LOAD(f28)
    LOAD(f29)
    LOAD(f30)
    LOAD(f31)
#undef LOAD

    // Tear down frame and retrieve the saved LR and CR from the caller's linkage area.
    masm.as_addi(StackPointer, StackPointer, (uint16_t)sizeof(EnterJITRegs));
    masm.as_ld(ScratchRegister, StackPointer, offsetof(EnterJITRegs, lr)); // caller
    masm.xs_mtlr(ScratchRegister);
    masm.as_ld(ScratchRegister, StackPointer, offsetof(EnterJITRegs, cr)); // caller
    masm.xs_mtcr(ScratchRegister);

    // Bye!
    masm.as_blr();
}

// static
mozilla::Maybe<::JS::ProfilingFrameIterator::RegisterState>
JitRuntime::getCppEntryRegisters(JitFrameLayout* frameStackAddress) {
  // Not supported, or not implemented yet.
  // TODO: Implement along with the corresponding stack-walker changes, in
  // coordination with the Gecko Profiler, see bug 1635987 and follow-ups.
  return mozilla::Nothing{};
}

void
JitRuntime::generateInvalidator(MacroAssembler& masm, Label* bailoutTail)
{
    AutoCreatedBy acb(masm, "JitRuntime::generateInvalidator");
    ADBlock("generateInvalidator");

    invalidatorOffset_ = startTrampolineCode(masm);

  // The InvalidationBailoutStack r3 points to must have:
  // - osiPointReturnAddress_
  // - ionScript_  (pushed by CodeGeneratorPPC64::generateInvalidateEpilogue())
  // - regs_  (pushed here)
  // - fpregs_  (pushed here) => r3

    // Stack has to be aligned here. If not, we will have to fix it.
    masm.checkStackAlignment();

    // Push registers such that we can access them from [base + code].
    masm.PushRegsInMask(AllRegs);

    // Pass pointer to InvalidationBailoutStack structure.
    masm.movePtr(StackPointer, r3);

    // Reserve place for BailoutInfo pointer. Two words to ensure alignment for
    // setupAlignedABICall.
    masm.subPtr(Imm32(2 * sizeof(uintptr_t)), StackPointer);
    // Pass pointer to BailoutInfo
    masm.movePtr(StackPointer, r4);

    using Fn = bool (*)(InvalidationBailoutStack * sp,
                        BaselineBailoutInfo * *info);
    masm.setupAlignedABICall();
    masm.passABIArg(r3);
    masm.passABIArg(r4);
    masm.callWithABI<Fn, InvalidationBailout>(
        ABIType::General, CheckUnsafeCallWithABI::DontCheckOther);

    masm.pop(r5);

    // Pop the machine state and the dead frame.
    masm.moveToStackPtr(FramePointer);

    // Jump to shared bailout tail.
    // The return code is left unchanged by this routine in r3.
    masm.jump(bailoutTail);
}

// XXX: completely rewritten, check in tests
void
JitRuntime::generateArgumentsRectifier(MacroAssembler& masm,
                                            ArgumentsRectifierKind kind)
{
    // Do not erase the frame pointer in this function.

    AutoCreatedBy acb(masm, "JitRuntime::generateArgumentsRectifier");
    ADBlock("generateArgumentsRectifier");

masm.xs_trap();

    // MIPS uses a5-a7, t0-t3 and s3, with s3 being the only callee-save reg.
    // We will do something similar for Power and use r4-r6, r7-r10 and r15.
    const Register nvRectReg = r15;

    const Register numArgsReg = r4;
    const Register numActArgsReg = r5;
    const Register calleeTokenReg = r6;
    const Register tempValue = r7;
    const Register numToPush = r8;
    const Register tempCalleeTokenReg = r9;
    const Register tempNumArgsReg = r10;

    switch (kind) {
        case ArgumentsRectifierKind::Normal:
            argumentsRectifierOffset_ = startTrampolineCode(masm);
            break;
        case ArgumentsRectifierKind::TrialInlining:
            trialInliningArgumentsRectifierOffset_ = startTrampolineCode(masm);
            break;
    }
    masm.pushReturnAddress();

    // Frame prologue.
    //
    // NOTE: if this changes, fix the Baseline bailout code too!
    // See BaselineStackBuilder::calculatePrevFramePtr and
    // BaselineStackBuilder::buildRectifierFrame (in BaselineBailouts.cpp).
    masm.push(FramePointer);
    masm.mov(StackPointer, FramePointer);

    // Load argc.
    masm.loadNumActualArgs(FramePointer, nvRectReg);
    // Load |nformals| into numArgsReg.
    masm.loadPtr(Address(FramePointer,
                 RectifierFrameLayout::offsetOfCalleeToken()),
                 calleeTokenReg);
    masm.mov(calleeTokenReg, numArgsReg);
    masm.andPtr(Imm32(uint32_t(CalleeTokenMask)), numArgsReg);
    masm.loadFunctionArgCount(numArgsReg, numArgsReg);

    // Stash another copy since we're going to clobber numArgsReg.
    masm.as_or(tempNumArgsReg, numArgsReg, numArgsReg);

    static_assert(CalleeToken_FunctionConstructing == 1,
      "Ensure that we can use the constructing bit to count the value");
    masm.mov(calleeTokenReg, tempCalleeTokenReg); // t2
    masm.ma_and(tempCalleeTokenReg, Imm32(uint32_t(CalleeToken_FunctionConstructing)));

  // Including |this|, and |new.target|, there are (|nformals| + 1 +
  // isConstructing) arguments to push to the stack.  Then we push a
  // JitFrameLayout.  We compute the padding expressed in the number of extra
  // |undefined| values to push on the stack.
  static_assert(
      sizeof(JitFrameLayout) % JitStackAlignment == 0,
      "No need to consider the JitFrameLayout for aligning the stack");
  static_assert(
      JitStackAlignment % sizeof(Value) == 0,
      "Ensure that we can pad the stack by pushing extra UndefinedValue");

  MOZ_ASSERT(mozilla::IsPowerOfTwo(JitStackValueAlignment));
  masm.add32(
      Imm32(JitStackValueAlignment - 1 /* for padding */ + 1 /* for |this| */),
      numArgsReg);
  masm.add32(tempCalleeTokenReg, numArgsReg);
  masm.and32(Imm32(~(JitStackValueAlignment - 1)), numArgsReg);

// s3 is nvRectReg and has argc

    // Load the number of |undefined|s to push (nargs - nvRectReg).
    masm.as_subf(numToPush, nvRectReg, numArgsReg); // T = B - A
    // ... and remove one for |this|
    masm.as_addi(numToPush, numToPush, -1);

    // Caller:
    // [arg2] [arg1] [this] [ [argc] [callee] [descr] [raddr] ] <- sp
    // '-nvRectReg-'
    //
    // Rectifier frame:
    // [fp'][undef] [undef] [undef] [arg2] [arg1] [this] [ [argc] [callee]
    //                                                   [descr] [raddr] ]
    //      '-------- r8 ---------' '-nvRectReg-'

    // Copy number of actual arguments into numActArgsReg.
    masm.mov(nvRectReg, numActArgsReg);

    masm.moveValue(UndefinedValue(), ValueOperand(tempValue));

    // Push undefined (including the padding).
    {
#if(0)
        Label undefLoopTop;

        masm.bind(&undefLoopTop);
        masm.sub32(Imm32(1), numToPush);
        masm.subPtr(Imm32(sizeof(Value)), StackPointer);
        masm.storeValue(ValueOperand(tempValue), Address(StackPointer, 0));

        masm.ma_bc(numToPush, numToPush, &undefLoopTop, Assembler::NonZero, ShortJump);
#else
        masm.xs_mtctr(numToPush);
        masm.as_stdu(tempValue, StackPointer, -sizeof(Value)); // -4
        masm.xs_bdnz(-4);
#endif
    }

    // Get the topmost argument.
    static_assert(sizeof(Value) == 8, "TimesEight is used to skip arguments");
    MOZ_ASSERT(tempValue == r7); // can clobber
    MOZ_ASSERT(numToPush == r8);
    MOZ_ASSERT(tempCalleeTokenReg == r9); // can clobber
    masm.x_slwi(r7, nvRectReg, 3); // r7 <- nargs * 8
    masm.as_add(numToPush, FramePointer, r7); // r8 <- fp + nargs * 8
    masm.addPtr(Imm32(sizeof(RectifierFrameLayout)), numToPush);

    // Push arguments, |nargs| + 1 times (to include |this|).
    masm.as_addi(nvRectReg, nvRectReg, 1);
    {
#if(0)
        Label copyLoopTop;

        masm.bind(&copyLoopTop);
        masm.sub32(Imm32(1), nvRectReg);
        masm.subPtr(Imm32(sizeof(Value)), StackPointer);
        masm.loadValue(Address(numToPush, 0), ValueOperand(tempValue));
        masm.storeValue(ValueOperand(tempValue), Address(StackPointer, 0));
        masm.subPtr(Imm32(sizeof(Value)), numToPush);

        masm.ma_bc(nvRectReg, nvRectReg, &copyLoopTop, Assembler::NonZero, ShortJump);
#else
        masm.xs_mtctr(nvRectReg);
        masm.as_ld(tempValue, numToPush, 0); // -12
        masm.as_addi(numToPush, numToPush, -8);
        masm.as_stdu(tempValue, StackPointer, -8);
        masm.xs_bdnz(-12);
#endif
    }

    // If constructing, copy newTarget also.
    {
        Label notConstructing;

        masm.branchTest32(Assembler::Zero, calleeTokenReg, Imm32(CalleeToken_FunctionConstructing),
                          &notConstructing);

        // thisFrame[numFormals] = prevFrame[argc]
        ValueOperand newTarget(tempValue);

        // Load vp[argc]. Add sizeof(Value) for |this|.
        BaseIndex newTargetSrc(FramePointer, numActArgsReg, TimesEight,
                           sizeof(RectifierFrameLayout) + sizeof(Value));
        masm.loadValue(newTargetSrc, newTarget);

        // Again, 1 for |this|. We bring back our saved register from above.
        BaseIndex newTargetDest(StackPointer, tempNumArgsReg, TimesEight, sizeof(Value));
        masm.storeValue(newTarget, newTargetDest);

        masm.bind(&notConstructing);
    }

    // Caller:
    // [arg2] [arg1] [this] [ [argc] [callee] [descr] [raddr] ]
    //
    //
    // Rectifier frame:
    // [fp'] <- fp [undef] [undef] [undef] [arg2] [arg1] [this] <- sp [ [argc]
    //                                              [callee] [descr] [raddr] ]

    // Construct JitFrameLayout.
    masm.push(calleeTokenReg);
    masm.pushFrameDescriptorForJitCall(FrameType::Rectifier, numActArgsReg,
                                       numActArgsReg);

    // Call the target function.
    masm.andPtr(Imm32(uint32_t(CalleeTokenMask)), calleeTokenReg);
    switch (kind) {
      case ArgumentsRectifierKind::Normal:
        masm.loadJitCodeRaw(calleeTokenReg, r8); // can clobber r8
        argumentsRectifierReturnOffset_ = masm.callJitNoProfiler(r8);
        break;
      case ArgumentsRectifierKind::TrialInlining:
        Label noBaselineScript, done;
        masm.loadBaselineJitCodeRaw(calleeTokenReg, r8, &noBaselineScript);
        masm.callJitNoProfiler(r8);
        masm.ma_b(&done, ShortJump);

        // See BaselineCacheIRCompiler::emitCallInlinedFunction.
        masm.bind(&noBaselineScript);
        masm.loadJitCodeRaw(calleeTokenReg, r8);
        masm.callJitNoProfiler(r8);
        masm.bind(&done);
        break;
    }

    masm.mov(FramePointer, StackPointer);
    masm.pop(FramePointer);
    masm.ret();
}

/* - When bailout is done via out of line code (lazy bailout).
 * Frame size is stored in LR (look at
 * CodeGeneratorPPC64::generateOutOfLineCode()) and thunk code should save it
 * on stack. In addition, snapshotOffset_ and padding_ are
 * pushed to the stack by CodeGeneratorPPC64::visitOutOfLineBailout().
 */
static void
PushBailoutFrame(MacroAssembler& masm, Register spArg)
{
    // Push the frameSize_ stored in LR.
    masm.xs_mflr(ScratchRegister);
    masm.push(ScratchRegister);

    // Push registers such that we can access them from [base + code].
    masm.PushRegsInMask(AllRegs);

    // Put pointer to BailoutStack as first argument to the Bailout().
    masm.movePtr(StackPointer, spArg);
}

static void
GenerateBailoutThunk(MacroAssembler& masm, Label* bailoutTail)
{
    PushBailoutFrame(masm, r3);

    // Put pointer to BailoutInfo.
    static const uint32_t sizeOfBailoutInfo = sizeof(uintptr_t); // * 2;
    masm.subPtr(Imm32(sizeOfBailoutInfo), StackPointer);
    masm.movePtr(StackPointer, r4);

    using Fn = bool (*)(BailoutStack * sp, BaselineBailoutInfo * *info);
    masm.setupUnalignedABICall(r5);
    masm.passABIArg(r3);
    masm.passABIArg(r4);
    masm.callWithABI<Fn, Bailout>(ABIType::General,
                                  CheckUnsafeCallWithABI::DontCheckOther);

    // Get BailoutInfo pointer.
    masm.loadPtr(Address(StackPointer, 0), r5);

    // Remove both the bailout frame and the topmost Ion frame's stack.
    masm.moveToStackPtr(FramePointer);

    // Jump to shared bailout tail. The BailoutInfo pointer is still in r5 and
    // the return code is already in r3, so we can just branch.
    masm.jump(bailoutTail);
}

void
JitRuntime::generateBailoutHandler(MacroAssembler& masm, Label* bailoutTail)
{
    AutoCreatedBy acb(masm, "JitRuntime::generateBailoutHandler");
    ADBlock("generateBailoutHandler");

    bailoutHandlerOffset_ = startTrampolineCode(masm);

    GenerateBailoutThunk(masm, bailoutTail);
}

bool
JitRuntime::generateVMWrapper(JSContext* cx, MacroAssembler& masm,
                              VMFunctionId id,
                              const VMFunctionData& f, DynFn nativeFun,
                              uint32_t* wrapperOffset)
{
    AutoCreatedBy acb(masm, "JitRuntime::generateVMWrapper");
    ADBlock("generateVMWrapper");

    *wrapperOffset = startTrampolineCode(masm);

    // Avoid conflicts with argument registers while discarding the result
    // after the function call.
    AllocatableGeneralRegisterSet regs(Register::Codes::WrapperMask);

    static_assert((Register::Codes::VolatileMask & ~Register::Codes::WrapperMask) == 0,
                  "Wrapper register set should be a superset of Volatile register set.");

    // The context is the first argument; r3 is the first argument register.
    Register cxreg = r3;
    regs.take(cxreg);

  // Even if this is a tail call, still push (see SharedICHelpers-ppc64-inl.h).
  // On link-register platforms, it is the responsibility of the VM *callee* to
  // push the return address, while the caller must ensure that the address
  // is stored in LR on entry. This allows the VM wrapper to work with both
  // direct calls and tail calls.
  masm.pushReturnAddress();


    // Push the frame pointer to finish the exit frame, then link it up.
    masm.Push(FramePointer);
  masm.moveStackPtrTo(FramePointer);
    masm.loadJSContext(cxreg);
    masm.enterExitFrame(cxreg, regs.getAny(), id);

  // Reserve space for the outparameter.
  masm.reserveVMFunctionOutParamSpace(f);
  
  masm.setupUnalignedABICallDontSaveRestoreSP();
  masm.passABIArg(cxreg);

  size_t argDisp = ExitFrameLayout::Size();

  // Copy any arguments.
  for (uint32_t explicitArg = 0; explicitArg < f.explicitArgs; explicitArg++) {
    switch (f.argProperties(explicitArg)) {
      case VMFunctionData::WordByValue:
        if (f.argPassedInFloatReg(explicitArg)) {
          masm.passABIArg(MoveOperand(FramePointer, argDisp), ABIType::Float64);
        } else {
          masm.passABIArg(MoveOperand(FramePointer, argDisp), ABIType::General);
        }
        argDisp += sizeof(void*);
        break;
      case VMFunctionData::WordByRef:
        masm.passABIArg(MoveOperand(FramePointer, argDisp,
                                    MoveOperand::Kind::EffectiveAddress),
                        ABIType::General);
        argDisp += sizeof(void*);
        break;
      case VMFunctionData::DoubleByValue:
      case VMFunctionData::DoubleByRef:
        MOZ_CRASH("NYI: riscv callVM should not be used with 128bits values.");
        break;
    }
  }

    // Copy the implicit outparam, if any.
    //if (InvalidReg != outReg)
     //   masm.passABIArg(outReg);
  // It is not a C++-abi outparam, which would get passed in the
  // outparam register, but a real parameter to the function, which
  // was stack-allocated above.
  const int32_t outParamOffset =
      -int32_t(ExitFooterFrame::Size()) - f.sizeOfOutParamStackSlot();
  if (f.outParam != Type_Void) {
    masm.passABIArg(MoveOperand(FramePointer, outParamOffset,
                                MoveOperand::Kind::EffectiveAddress),
                    ABIType::General);
  }

    masm.callWithABI(nativeFun, ABIType::General,
                     CheckUnsafeCallWithABI::DontCheckHasExitFrame);

    // Test for failure.
    switch (f.failType()) {
      case Type_Cell:
        masm.branchTestPtr(Assembler::Zero, r3, r3, masm.failureLabel());
        break;
      case Type_Bool:
        // Called functions return bools, which are 0/false and non-zero/true.
        masm.branchIfFalseBool(r3, masm.failureLabel());
        break;
      case Type_Void:
        break;
      default:
        MOZ_CRASH("unknown failure kind");
    }

  // Load the outparam. Leave the cannoli.
  masm.loadVMFunctionOutParam(f, Address(FramePointer, outParamOffset));

  // Pop frame and restore frame pointer.
  masm.moveToStackPtr(FramePointer);
  masm.pop(FramePointer);

  // Return. Subtract sizeof(void*) for the frame pointer.
  masm.retn(Imm32(sizeof(ExitFrameLayout) - sizeof(void*) +
                  f.explicitStackSlots() * sizeof(void*) +
                  f.extraValuesToPop * sizeof(Value)));

    return true;
}

uint32_t
JitRuntime::generatePreBarrier(JSContext* cx, MacroAssembler& masm, MIRType type)
{
    AutoCreatedBy acb(masm, "JitRuntime::generatePreBarrier");
    ADBlock("generatePreBarrier");

    uint32_t offset = startTrampolineCode(masm);

    MOZ_ASSERT(PreBarrierReg == r4);
    Register temp1 = r3;
    Register temp2 = r5;
    Register temp3 = r6;
    // TODO: could be more efficient with multipush/pop
    masm.push(temp1);
    masm.push(temp2);
    masm.push(temp3);

    Label noBarrier;
    masm.emitPreBarrierFastPath(cx->runtime(), type, temp1, temp2, temp3, &noBarrier);

    // Call into C++ to mark this GC thing.
    masm.pop(temp3);
    masm.pop(temp2);
    masm.pop(temp1);

    // Explicitly save LR, since we can't use it in PushRegsInMask.
    masm.xs_mflr(ScratchRegister);
    masm.push(ScratchRegister);
    LiveRegisterSet save;
    save.set() = RegisterSet(GeneralRegisterSet(Registers::VolatileMask),
                             FloatRegisterSet(FloatRegisters::VolatileMask));
    masm.PushRegsInMask(save);

    masm.movePtr(ImmPtr(cx->runtime()), r3);

    masm.setupUnalignedABICall(r5);
    masm.passABIArg(r3);
    masm.passABIArg(r4);
    masm.callWithABI(JitPreWriteBarrier(type));

    masm.PopRegsInMask(save);
    // ret() pops LR for us.
    masm.ret();

    masm.bind(&noBarrier);
    masm.pop(temp3);
    masm.pop(temp2);
    masm.pop(temp1);
    masm.abiret();

    return offset;
}

void
JitRuntime::generateBailoutTailStub(MacroAssembler& masm, Label* bailoutTail)
{
    AutoCreatedBy acb(masm, "JitRuntime::generateBailoutTailStub");
    ADBlock("generateBailoutTailStub");

    masm.bind(bailoutTail);
    masm.generateBailoutTail(r4, r5);
}

